@@ -21,7 +21,6 @@
 #include <arpa/inet.h>
 #include <sys/ioctl.h>
 #include <bpf/libbpf.h>
-#include <sys/sysinfo.h>
 #include <sys/timerfd.h>
 #include <sys/utsname.h>
 #include <linux/limits.h>
@@ -1480,16 +1479,10 @@ static void swap(struct stats_record **a, struct stats_record **b)
 	*b = tmp;
 }
 
-static int sample_timer_cb(int timerfd, struct stats_record **rec,
-			   struct stats_record **prev)
+static int print_stats(struct stats_record **rec, struct stats_record **prev)
 {
 	char line[64] = "Summary";
 	int ret;
-	__u64 t;
-
-	ret = read(timerfd, &t, sizeof(t));
-	if (ret < 0)
-		return -errno;
 
 	swap(prev, rec);
 	ret = sample_stats_collect(*rec);
@@ -1514,6 +1507,19 @@ static int sample_timer_cb(int timerfd, struct stats_record **rec,
 	return 0;
 }
 
+static int sample_timer_cb(int timerfd, struct stats_record **rec,
+			   struct stats_record **prev)
+{
+	int ret;
+	__u64 t;
+
+	ret = read(timerfd, &t, sizeof(t));
+	if (ret < 0)
+		return -errno;
+
+	return print_stats(rec, prev);
+}
+
 int sample_run(int interval, void (*post_cb)(void *), void *ctx)
 {
 	struct timespec ts = { interval, 0 };
@@ -1573,9 +1579,11 @@ int sample_run(int interval, void (*post_cb)(void *), void *ctx)
 				break;
 		}
 
-		if (pfd[0].revents & POLLIN)
+		if (pfd[0].revents & POLLIN) {
 			ret = sample_signal_cb();
-		else if (pfd[1].revents & POLLIN)
+			if (ret)
+				print_stats(&rec, &prev);
+		} else if (pfd[1].revents & POLLIN)
 			ret = sample_timer_cb(timerfd, &rec, &prev);
 
 		if (ret)
